unit Main;

interface

uses
  System.SysUtils, System.Types, System.UITypes, System.Classes, System.Variants,
  FMX.Types, FMX.Controls, FMX.Forms, FMX.Graphics, FMX.Dialogs,
  FMX.Controls.Presentation, FMX.StdCtrls, FMX.ScrollBox, FMX.Memo,
  iOSapi.Foundation, FMX.TabControl, FMX.Memo.Types,
  Winsoft.iOS.CoreNFC,  Winsoft.iOS.Nfc, Winsoft.iOS.NfcDna;

type
  TFormMain = class(TForm)
    ButtonTest: TButton;
    MemoTest: TMemo;
    TabControl: TTabControl;
    TabItemTest: TTabItem;
    TabItemReadEV2: TTabItem;
    TabItemWriteEV2: TTabItem;
    TabItemReadLRP: TTabItem;
    TabItemWriteLRP: TTabItem;
    MemoReadEV2: TMemo;
    ButtonReadEV2: TButton;
    MemoWriteEV2: TMemo;
    ButtonWriteEV2: TButton;
    MemoReadLRP: TMemo;
    ButtonReadLRP: TButton;
    MemoWriteLRP: TMemo;
    ButtonWriteLRP: TButton;
    procedure ButtonTestClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
  private
    { Private declarations }
    NfcTagSession: TNfcTagSession;
    NTag424: TNTag424;
    function GetCurrentMemo: TMemo;
    procedure ClearLines;
    procedure AddLine(const Text: string = '');
    procedure ShowError(const Error: TNfcError);
    function CheckException(const E: Exception): Boolean;
    procedure NfcActive(Sender: TObject);
    procedure NfcDetect(Sender: TObject; const Tags: TNfcTags);
    procedure NfcError(Sender: TObject; const Error: TNfcError);
    procedure ReadEV2(const ISO7816Tag: TISO7816Tag);
    procedure WriteEV2(const ISO7816Tag: TISO7816Tag);
    procedure ReadLRP(const ISO7816Tag: TISO7816Tag);
    procedure WriteLRP(const ISO7816Tag: TISO7816Tag);
    procedure Test(const ISO7816Tag: TISO7816Tag);
  public
    { Public declarations }
  end;

var
  FormMain: TFormMain;

implementation

uses Winsoft.iOS.Ev2;

{$R *.fmx}

function ToStr(Communication: TCommunication): string; overload;
begin
  case Communication of
    TCommunication.Plain: Result := 'Plain';
    TCommunication.Mac:   Result := 'Mac';
    TCommunication.Full:  Result := 'Full';
  end;
end;

function ToStr(Access: TAccess): string; overload;
begin
  case Access of
    TAccess.MasterKey: Result := 'Master key';
    TAccess.AppKey1:   Result := 'App key 1';
    TAccess.AppKey2:   Result := 'App key 2';
    TAccess.AppKey3:   Result := 'App key 3';
    TAccess.AppKey4:   Result := 'App key 4';
    TAccess.Free:      Result := 'Free access';
    TAccess.No:        Result := 'No access';
  end;
end;

function TFormMain.GetCurrentMemo: TMemo;
begin
  if TabControl.ActiveTab = TabItemReadEV2 then
    Result := MemoReadEV2
  else if TabControl.ActiveTab = TabItemWriteEV2 then
    Result := MemoWriteEV2
  else if TabControl.ActiveTab = TabItemReadLRP then
    Result := MemoReadLRP
  else if TabControl.ActiveTab = TabItemWriteLRP then
    Result := MemoWriteLRP
  else
    Result := MemoTest
end;

procedure TFormMain.ClearLines;
var Memo: TMemo;
begin
  Memo := GetCurrentMemo;
  Memo.SetFocus;
  Memo.Lines.Clear;
  Memo.Repaint;
end;

procedure TFormMain.AddLine(const Text: string);
var Memo: TMemo;
begin
  Memo := GetCurrentMemo;
  Memo.SetFocus;
  Memo.Lines.Add(Text);
  Memo.Repaint;
end;

procedure TFormMain.ShowError(const Error: TNfcError);
begin
  with Error do
    if IsError then
    begin
      AddLine('Error');
      AddLine('  Domain: ' + Domain);
      AddLine('  Code: ' + IntToStr(Code));
      AddLine('  Description: ' + Description);
      AddLine('  Failure reason: ' + FailureReason);
      AddLine('  Recovery suggestion: ' + RecoverySuggestion);
      AddLine;
    end;
end;

procedure TFormMain.NfcActive(Sender: TObject);
begin
  AddLine('Active event');
end;

function TFormMain.CheckException(const E: Exception): Boolean;
begin
  Result := E <> nil;
  if Result then
    NfcTagSession.Invalidate('Error: ' + E.Message);
end;

procedure TFormMain.NfcDetect(Sender: TObject; const Tags: TNfcTags);
var
  I: Integer;
  ISO7816Tag: TISO7816Tag;
begin
  try
    ClearLines;
    AddLine('TAG detected: ' + IntToStr(Length(Tags)) + ' tags');
    for I := 0 to Length(Tags) - 1 do
      case Tags[I].TagType of
        TTagType.ISO7816Compatible:
        begin
          ISO7816Tag := Tags[I].ISO7816Tag;
          AddLine('ISO7816 TAG');
          AddLine('AID: ' + ISO7816Tag.AID);
          AddLine('Application data: ' + BytesToHex(ISO7816Tag.ApplicationData));
          AddLine('Historical bytes: ' + BytesToHex(ISO7816Tag.HistoricalBytes));
          AddLine('Identifier: ' + BytesToHex(ISO7816Tag.Identifier));
          AddLine('Proprietary coding: ' + BoolToStr(ISO7816Tag.ProprietaryCoding, True));
          AddLine;

          NfcTagSession.Connect(Tags[I],
            procedure (const Error: TNfcError)
            begin
              if Error.IsError then
              begin
                ShowError(Error);
                NfcTagSession.Invalidate('Error: ' + Error.Description);
              end
              else
              begin
                if TabControl.ActiveTab = TabItemReadEV2 then
                  ReadEV2(ISO7816Tag)
                else if TabControl.ActiveTab = TabItemWriteEV2 then
                  WriteEV2(ISO7816Tag)
                else if TabControl.ActiveTab = TabItemReadLRP then
                  ReadLRP(ISO7816Tag)
                else if TabControl.ActiveTab = TabItemWriteLRP then
                  WriteLRP(ISO7816Tag)
                else
                  Test(ISO7816Tag)
              end;
            end);
        end;

        else
          AddLine('Unsupported tag type');
      end;

    NfcTagSession.Invalidate;
  except
    on E: Exception do
      AddLine('Exception: ' + E.Message);
  end;
end;

procedure TFormMain.ReadEV2(const ISO7816Tag: TISO7816Tag);
begin
  NTag424.Free;
  NTag424 := TNTag424.Create(ISO7816Tag);

  AddLine('IsoSelectDedicatedFile');
  NTag424.IsoSelectDedicatedFile(
    procedure (const E: Exception)
    begin
      if not CheckException(E) then
      begin
        AddLine('Authenticate EV2 with App Key 3');
        NTag424.Authenticate(TAuthentication.Ev2, TKey.Key3, Block00,
          procedure (const E: Exception)
          begin
            if not CheckException(E) then
            begin
              AddLine('Transaction ID: $' + IntToHex(NTag424.TransactionId, 8));

              // read File3 content
              NTag424.ReadData(TFile.File3, TCommunication.Full,
                procedure (const Data: TBytes; const E: Exception)
                begin
                  if not CheckException(E) then
                  begin
                    AddLine('File3 content: $' + BytesToHex(Data));
                    NfcTagSession.Invalidate;
                  end;
                end);
            end;
          end);
      end;
    end);
end;

procedure TFormMain.WriteEV2(const ISO7816Tag: TISO7816Tag);
begin
  NTag424.Free;
  NTag424 := TNTag424.Create(ISO7816Tag);

  AddLine('IsoSelectDedicatedFile');
  NTag424.IsoSelectDedicatedFile(
    procedure (const E: Exception)
    begin
      if not CheckException(E) then
      begin
        AddLine('Authenticate EV2 with App Key 3');
        NTag424.Authenticate(TAuthentication.Ev2, TKey.Key3, Block00,
          procedure (const E: Exception)
          var Data: TBytes;
          begin
            if not CheckException(E) then
            begin
              AddLine('Transaction ID: $' + IntToHex(NTag424.TransactionId, 8));

              // write File3 content
              Data := TBytes.Create(1, 2, 3, 4, 5, 6, 7, 8, 9, 10);
              AddLine('File3 write data at offset 5: $' + BytesToHex(Data));
              NTag424.WriteData(TFile.File3, TCommunication.Full, Data,
                procedure (const E: Exception)
                begin
                  if not CheckException(E) then
                  begin
                    AddLine('File3 content written');
                    NfcTagSession.Invalidate;
                  end;
                end, 5);
            end;
          end);
      end;
    end);
end;

procedure TFormMain.ReadLRP(const ISO7816Tag: TISO7816Tag);
begin
  NTag424.Free;
  NTag424 := TNTag424.Create(ISO7816Tag);

  AddLine('IsoSelectDedicatedFile');
  NTag424.IsoSelectDedicatedFile(
    procedure (const E: Exception)
    begin
      if not CheckException(E) then
      begin
        AddLine('Authenticate LRP with App Key 3');
        NTag424.Authenticate(TAuthentication.Lrp, TKey.Key3, Block00,
          procedure (const E: Exception)
          begin
            if not CheckException(E) then
            begin
              AddLine('Transaction ID: $' + IntToHex(NTag424.TransactionId, 8));

              // read File3 content
              NTag424.ReadData(TFile.File3, TCommunication.Full,
                procedure (const Data: TBytes; const E: Exception)
                begin
                  if not CheckException(E) then
                  begin
                    AddLine('File3 content: $' + BytesToHex(Data));
                    NfcTagSession.Invalidate;
                  end;
                end);
            end;
          end);
      end;
    end);
end;

procedure TFormMain.WriteLRP(const ISO7816Tag: TISO7816Tag);
begin
  NTag424.Free;
  NTag424 := TNTag424.Create(ISO7816Tag);

  AddLine('IsoSelectDedicatedFile');
  NTag424.IsoSelectDedicatedFile(
    procedure (const E: Exception)
    begin
      if not CheckException(E) then
      begin
        AddLine('Authenticate LRP with App Key 3');
        NTag424.Authenticate(TAuthentication.Lrp, TKey.Key3, Block00,
          procedure (const E: Exception)
          var Data: TBytes;
          begin
            if not CheckException(E) then
            begin
              AddLine('Transaction ID: $' + IntToHex(NTag424.TransactionId, 8));

              // write File3 content
              Data := TBytes.Create(1, 2, 3, 4, 5, 6, 7, 8, 9, 10);
              AddLine('File3 write data at offset 5: $' + BytesToHex(Data));
              NTag424.WriteData(TFile.File3, TCommunication.Full, Data,
                procedure (const E: Exception)
                begin
                  if not CheckException(E) then
                  begin
                    AddLine('File3 content written');
                    NfcTagSession.Invalidate;
                  end;
                end, 5);
            end;
          end);
      end;
    end);
end;

procedure TFormMain.Test(const ISO7816Tag: TISO7816Tag);
begin
  NTag424.Free;
  NTag424 := TNTag424.Create(ISO7816Tag);

  NTag424.GetVersion(
    procedure (const Data: TBytes; const E: Exception)
    begin
      if not CheckException(E) then
      begin
        AddLine('Version: $' + BytesToHex(Data));
        NfcTagSession.Invalidate;
      end;
    end);

{
  NTag424.GetKeyVersion(TKey.Key1, // Originality key 1
    procedure (Data: Byte; const E: Exception)
    begin
      if not CheckException(E) then
      begin
        AddLine('Key1Version: $' + IntToHex(Data, 2));
        NfcTagSession.Invalidate;
      end;
    end);
}

{$ifdef UNCOMMENT}
  AddLine('IsoSelectDedicatedFile');
  NTag424.IsoSelectDedicatedFile(
    procedure (const E: Exception)
    begin
      if not CheckException(E) then
      begin
{
        NTag424.GetKeyVersion(TKey.MasterKey, // Master key
          procedure (Data: Byte; const E: Exception)
          begin
            if not CheckException(E) then
            begin
              AddLine('MasterKeyVersion: $' + IntToHex(Data, 2));
              NfcTagSession.Invalidate;
            end;
          end);
}

(*
        AddLine('IsoSelect(File2)');
        NTag424.IsoSelect(TFile.File2,
          procedure (const E: Exception)
          var Data: TBytes;
          begin
            if not CheckException(E) then
            begin
{
              NTag424.IsoReadBinary(
                procedure (const Data: TBytes; const E: Exception)
                begin
                  if not CheckException(E) then
                  begin
                    AddLine('File content: $' + BytesToHex(Data));
                    NfcTagSession.Invalidate;
                  end;
                end);
}
{
              Data := TBytes.Create(1, 2, 3, 4, 5, 6, 7, 8, 9, 10);
              NTag424.IsoUpdateBinary(Data, 10,
                procedure (const E: Exception)
                begin
                  if not CheckException(E) then
                  begin
                    AddLine('File updated');
                    NfcTagSession.Invalidate;
                  end;
                end);
}
            end;
          end);
*)

(*
        AddLine('Authenticate EV2 with App Key 3');
        NTag424.Authenticate(TAuthentication.Ev2, TKey.Key3, Block00,
          procedure (const E: Exception)
          begin
            if not CheckException(E) then
            begin
              AddLine('Transaction ID: $' + IntToHex(NTag424.TransactionId, 8));

{
              NTag424.GetCardUid(
                procedure (const Data: TBytes; const E: Exception)
                begin
                  if not CheckException(E) then
                  begin
                    AddLine('Card UID: $' + BytesToHex(Data));
                    NfcTagSession.Invalidate;
                  end;
                end);
}

{
              NTag424.GetFileSettings(TFile.File1,
                procedure (Mirroring: Boolean; Communication: TCommunication;
                  Read, Write, ReadWrite, ChangeSettings: TAccess; FileSize: Integer; const E: Exception)
                begin
                  if not CheckException(E) then
                  begin
                    AddLine('File1 mirroring: ' + BoolToStr(Mirroring, True));
                    AddLine('File1 communication: ' + ToStr(Communication));
                    AddLine('File1 read access: ' + ToStr(Read));
                    AddLine('File1 write access: ' + ToStr(Write));
                    AddLine('File1 read/write access: ' + ToStr(ReadWrite));
                    AddLine('File1 change settings access: ' + ToStr(ChangeSettings));
                    AddLine('File1 size: ' + IntToStr(FileSize) + ' bytes');
                    NfcTagSession.Invalidate;
                  end;
                end);
}

{
              NTag424.GetFileSettings(TFile.File2,
                procedure (Mirroring: Boolean; Communication: TCommunication;
                  Read, Write, ReadWrite, ChangeSettings: TAccess; FileSize: Integer; const E: Exception)
                begin
                  if not CheckException(E) then
                  begin
                    AddLine('File2 mirroring: ' + BoolToStr(Mirroring, True));
                    AddLine('File2 communication: ' + ToStr(Communication));
                    AddLine('File2 read access: ' + ToStr(Read));
                    AddLine('File2 write access: ' + ToStr(Write));
                    AddLine('File2 read/write access: ' + ToStr(ReadWrite));
                    AddLine('File2 change settings access: ' + ToStr(ChangeSettings));
                    AddLine('File2 size: ' + IntToStr(FileSize) + ' bytes');
                    NfcTagSession.Invalidate;
                  end;
                end);
}

{
              AddLine('File2 change settings');
              NTag424.ChangeFileSettings(TFile.File2, TCommunication.Plain, TAccess.Free, TAccess.Free, TAccess.Free, TAccess.MasterKey,
                procedure (const E: Exception)
                begin
                  if not CheckException(E) then
                  begin
                     NfcTagSession.Invalidate;
                  end;
                end);
}

{
              NTag424.ChangeKey(TKey.MasterKey, Block00, Block00, $14,
                procedure (const E: Exception)
                begin
                  if not CheckException(E) then
                  begin

                    NTag424.GetKeyVersion(TKey.MasterKey, // Master key
                      procedure (Data: Byte; const E: Exception)
                      begin
                        if not CheckException(E) then
                        begin
                          AddLine('MasterKeyVersion: $' + IntToHex(Data, 2));
                          NfcTagSession.Invalidate;
                        end;
                      end);

                     // NfcTagSession.Invalidate;
                  end;
                end);
}

{
              NTag424.ReadSignature(
                procedure (const Data: TBytes; const E: Exception)
                begin
                  if not CheckException(E) then
                  begin
                    AddLine('Signature: $' + BytesToHex(Data));
                    NfcTagSession.Invalidate;
                  end;
                end);
}
              // NfcTagSession.Invalidate;
            end;
          end);
*)
      end;
    end);
{$endif UNCOMMENT}
end;

procedure TFormMain.NfcError(Sender: TObject; const Error: TNfcError);
begin
  if Error.Code <> NFCReaderSessionInvalidationErrorUserCanceled then
  begin
    AddLine('Error event');
    ShowError(Error);
  end;
end;

procedure TFormMain.ButtonTestClick(Sender: TObject);
begin
  ClearLines;
  try
    if not TNfcTagSession.ReadingAvailable then
      AddLine('NFC is not available on this device')
    else
    begin
      NfcTagSession.Free;
      NfcTagSession := TNfcTagSession.Create;
      NfcTagSession.AlertMessage := 'Hold your iPhone near the item.';
      NfcTagSession.OnActive := NfcActive;
      NfcTagSession.OnDetect := NfcDetect;
      NfcTagSession.OnError := NfcError;
      NfcTagSession.Scan;
    end;
  except
    on E: Exception do
      AddLine('Exception: ' + E.Message);
  end;
end;

procedure TFormMain.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  NTag424.Free;
  NfcTagSession.Free;
end;

end.
