unit Main;

interface

uses
  System.SysUtils, System.Types, System.UITypes, System.Classes, System.Variants,
  FMX.Types, FMX.Controls, FMX.Forms, FMX.Graphics, FMX.Dialogs,
  FMX.Controls.Presentation, FMX.StdCtrls, FMX.ScrollBox, FMX.Memo,
  iOSapi.Foundation, FMX.TabControl, FMX.Memo.Types,
  Winsoft.iOS.CoreNFC, Winsoft.iOS.Nfc;

type
  TFormMain = class(TForm)
    ButtonReadNdef: TButton;
    MemoReadNdef: TMemo;
    TabControl: TTabControl;
    TabItemReadNdef: TTabItem;
    TabItemWriteNdef: TTabItem;
    ButtonWriteNdef: TButton;
    MemoWriteNdef: TMemo;
    TabItemReadTag: TTabItem;
    TabItemWriteTag: TTabItem;
    ButtonReadTag: TButton;
    MemoReadTag: TMemo;
    ButtonWriteTag: TButton;
    MemoWriteTag: TMemo;
    procedure ButtonNdefClick(Sender: TObject);
    procedure ButtonNfcTagClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
  private
    { Private declarations }
    NDefSession: TNDefSession;
    NfcTagSession: TNfcTagSession;
    function Memo: TMemo;
    procedure AddLine(const Text: string = '');
    procedure ShowError(const Error: TNfcError);
    procedure NfcActive(Sender: TObject);
    procedure NDefDetect(Sender: TObject; const Tags: TNDefTags);
    function NDefError(const Error: TNfcError): Boolean;
    procedure NfcDetect(Sender: TObject; const Tags: TNfcTags);
    function NfcError(const Error: TNfcError): Boolean; overload;
    procedure NfcError(Sender: TObject; const Error: TNfcError); overload;
    procedure ReadTag(const Tag: TNDefTag);
    procedure WriteTag(const Tag: TNDefTag);
  public
    { Public declarations }
  end;

var
  FormMain: TFormMain;

implementation

{$R *.fmx}

function ToStr(Format: TNDefFormat): string; overload;
begin
  case Format of
    TNDefFormat.Empty:         Result := 'Empty';
    TNDefFormat.WellKnown:     Result := 'Well known';
    TNDefFormat.Media:         Result := 'Media';
    TNDefFormat.AbsoluteUri:   Result := 'Absolute URI';
    TNDefFormat.External:      Result := 'External';
    TNDefFormat.Unchanged:     Result := 'Unchanged';
    else {TNDefFormat.Unknown} Result := 'Unknown';
  end;
end;

function ToStr(Status: TNDefStatus): string; overload;
begin
  case Status of
    TNDefStatus.NotSupported: Result := 'Not supported';
    TNDefStatus.ReadWrite:    Result := 'Read/write';
    TNDefStatus.ReadOnly:     Result := 'Read only';
    else                      Result := 'Unknown';
  end;
end;

function ToStr(Family: TMiFareFamily): string; overload;
begin
  case Family of
    TMiFareFamily.Ultralight: Result := 'Ultralight';
    TMiFareFamily.Plus:       Result := 'Plus';
    TMiFareFamily.DESFire:    Result := 'DESFire';
    else                      Result := 'Unknown';
  end;
end;

function TFormMain.Memo: TMemo;
begin
  if TabControl.ActiveTab = TabItemReadNdef then
    Result := MemoReadNdef
  else if TabControl.ActiveTab = TabItemWriteNdef then
    Result := MemoWriteNdef
  else if TabControl.ActiveTab = TabItemReadTag then
    Result := MemoReadTag
  else if TabControl.ActiveTab = TabItemWriteTag then
    Result := MemoWriteTag
  else
    Result := nil;
end;

procedure TFormMain.AddLine(const Text: string);
begin
  Memo.SetFocus;
  Memo.Lines.Add(Text);
  Memo.Repaint;
end;

procedure TFormMain.ShowError(const Error: TNfcError);
begin
  with Error do
    if IsError then
    begin
      AddLine('Error');
      AddLine('  Domain: ' + Domain);
      AddLine('  Code: ' + IntToStr(Code));
      AddLine('  Description: ' + Description);
      AddLine('  Failure reason: ' + FailureReason);
      AddLine('  Recovery suggestion: ' + RecoverySuggestion);
      AddLine;
    end;
end;

procedure TFormMain.NfcActive(Sender: TObject);
begin
  AddLine('Active event');
end;

function TFormMain.NDefError(const Error: TNfcError): Boolean;
begin
  Result := Error.IsError;
  if Result then
  begin
    ShowError(Error);
    NDefSession.Invalidate('Error: ' + Error.Description);
  end
end;

procedure TFormMain.NDefDetect(Sender: TObject; const Tags: TNDefTags);
var
  I: Integer;
  Tag: TNDefTag;
begin
  try
    AddLine('NDEF detected: ' + IntToStr(Length(Tags)) + ' tags');
    for I := 0 to Length(Tags) - 1 do
    begin
      Tag := Tags[I];

      // connect to NDEF tag
      NDefSession.Connect(Tag,
        procedure (const Error: TNfcError)
        begin
          if not NDefError(Error) then
            if not Tag.IsAvailable then
            begin
              AddLine('Tag not available');
              NDefSession.Invalidate('Tag not available');
            end
            else
              // retrieve NDEF tag info
              Tag.QueryStatus(
                procedure (Status: TNDefStatus; Capacity: LongWord; const Error: TNfcError)
                begin
                  if not NDefError(Error) then
                  begin
                    AddLine('Status: ' + ToStr(Status));
                    AddLine('Capacity: ' + IntToStr(Capacity));

                    // READ/WRITE tag
                    if TabControl.ActiveTab = TabItemReadNdef then
                      ReadTag(Tag)
                    else
                      WriteTag(Tag);
                  end;
                end);
        end);
    end;

    NDefSession.Invalidate;
  except
    on E: Exception do
      AddLine('Exception: ' + E.Message);
  end;
end;

procedure TFormMain.ReadTag(const Tag: TNDefTag);
begin
  Tag.Read(
    procedure (const Message: TNDefMessage; const Error: TNfcError)
    var
      I: Integer;
      Payload: TNDefPayload;
      PayloadType, PayloadData: TBytes;
    begin
      AddLine('Read');
      if not NDefError(Error) then
      begin
        AddLine('NDEF Message');
        for I := 0 to Message.PayloadCount - 1 do
        begin
          Payload := Message.Payloads[I];
          AddLine('  Identifier: ' + BytesToHex(Payload.Identifier));
          AddLine('  Format: ' + ToStr(Payload.Format));

          PayloadType := Payload.&Type;
          if Payload.Format = TNDefFormat.WellKnown then
            AddLine('  Type: ' + BytesToHex(PayloadType) + ' ' + TEncoding.UTF8.GetString(PayloadType))
          else
            AddLine('  Type: ' + BytesToHex(PayloadType));

          PayloadData := Payload.Data;
          AddLine('  Data: ' + BytesToHex(PayloadData));
          if Payload.Format = TNDefFormat.WellKnown then
          begin
            if TEncoding.UTF8.GetString(PayloadType) = 'T' then
              AddLine('  Text: ' + TEncoding.UTF8.GetString(PayloadData, 3, Length(PayloadData) - 3))
            else if TEncoding.UTF8.GetString(PayloadType) = 'U' then
              AddLine('  URL data: ' + TEncoding.UTF8.GetString(PayloadData, 1, Length(PayloadData) - 1));
          end;
          AddLine;
        end;

        NDefSession.Invalidate;
      end;
    end);
end;

procedure TFormMain.WriteTag(const Tag: TNDefTag);
begin
  Tag.Write(
    TNDefMessage.Create(
    [
      // Text record
      TNDefPayload.Create(TNDefFormat.WellKnown, TEncoding.UTF8.GetBytes('T'), nil, TEncoding.UTF8.GetBytes(#$02#$65#$6E + 'Hello, world!')),
      // URI record
      TNDefPayload.Create(TNDefFormat.WellKnown, TEncoding.UTF8.GetBytes('U'), nil, TEncoding.UTF8.GetBytes(#$02 + 'winsoft.sk'))
    ]),
    procedure (const Error: TNfcError)
    begin
      AddLine('Write');
      if not NDefError(Error) then
        NDefSession.Invalidate;
    end);
end;

function TFormMain.NfcError(const Error: TNfcError): Boolean;
begin
  Result := Error.IsError;
  if Result then
  begin
    ShowError(Error);
    NfcTagSession.Invalidate('Error: ' + Error.Description);
  end
end;

procedure TFormMain.NfcDetect(Sender: TObject; const Tags: TNfcTags);
var
  I: Integer;
  ISO15693Tag: TISO15693Tag;
  ISO7816Tag: TISO7816Tag;
  FeliCaTag: TFeliCaTag;
  MiFareTag: TMiFareTag;
begin
  try
    AddLine('TAG detected: ' + IntToStr(Length(Tags)) + ' tags');
    for I := 0 to Length(Tags) - 1 do
      case Tags[I].TagType of

        TTagType.ISO15693:
        begin
          ISO15693Tag := Tags[I].ISO15693Tag;
          AddLine('ISO15693 TAG');
          AddLine('Identifier: ' + BytesToHex(ISO15693Tag.Identifier));
          AddLine('Manufacturer: $' + IntToHex(ISO15693Tag.Manufacturer));
          AddLine('Serial number: ' + BytesToHex(ISO15693Tag.SerialNumber));
          AddLine;

          NfcTagSession.Connect(Tags[I],
            procedure (const Error: TNfcError)
            begin
              if not NfcError(Error) then
                ISO15693Tag.GetInfo(
                  procedure (DsfId, Afi, BlockSize, BlockCount: Integer)
                  begin
                    if BlockCount < 0 then
                      NfcTagSession.Invalidate('Error')
                    else
                    begin
                      AddLine('AFI: $' + IntToHex(Afi, 2));
                      AddLine('DSF ID: $' + IntToHex(DsfId, 2));
                      AddLine('Block size: ' + IntToStr(BlockSize));
                      AddLine('Block count: ' + IntToStr(BlockCount));
                      AddLine;

                      ISO15693Tag.ReadBlocks(0, BlockCount,
                        procedure (const Data: TArray<TBytes>; const Error: TNfcError)
                        var I: Integer;
                        begin
                          if not NfcError(Error) then
                          begin
                            for I := 0 to Length(Data) - 1 do
                              AddLine('Block $' + IntToHex(I, 2) + ': ' + BytesToHex(Data[I]));

                            NfcTagSession.Invalidate;
                          end;
                        end);
                    end;
                  end);
            end);
        end;

        TTagType.ISO7816Compatible:
        begin
          ISO7816Tag := Tags[I].ISO7816Tag;
          AddLine('ISO7816 TAG');
          AddLine('AID: ' + ISO7816Tag.AID);
          AddLine('Application data: ' + BytesToHex(ISO7816Tag.ApplicationData));
          AddLine('Historical bytes: ' + BytesToHex(ISO7816Tag.HistoricalBytes));
          AddLine('Identifier: ' + BytesToHex(ISO7816Tag.Identifier));
          AddLine;
        end;

        TTagType.FeliCa:
        begin
          FeliCaTag := Tags[I].FeliCaTag;
          AddLine('FeliCa TAG');
          AddLine('Manufacturer: ' + BytesToHex(FeliCaTag.Manufacturer));
          AddLine('System code: ' + BytesToHex(FeliCaTag.SystemCode));
          AddLine;
        end;

        TTagType.MiFare:
        begin
          MiFareTag := Tags[I].MiFareTag;
          AddLine('MiFare TAG');
          AddLine('Family: ' + ToStr(MiFareTag.Family));
          AddLine('Historical bytes: ' + BytesToHex(MiFareTag.HistoricalBytes));
          AddLine('Identifier: ' + BytesToHex(MiFareTag.Identifier));
          AddLine;

          NfcTagSession.Connect(Tags[I],
            procedure (const Error: TNfcError)
            const
              PageCount = 44;
            type
              TReadProc = reference to procedure (Page: Byte);
            var
              Page: Byte;
              ReadProc: TReadProc;
            begin
              if not NfcError(Error) then
              begin
                for Page := 0 to PageCount div 4 - 1 do
                begin
                  ReadProc :=
                    procedure (Page: Byte) // Page variable has to be captured
                    begin
                      MiFareTag.Read(Page,
                        procedure (const Response: TBytes; const Error: TNfcError)
                        begin
                          if Error.IsError then
                            AddLine('Page $' + IntToHex(Page, 2) + ': Error ' + Error.Description)
                          else
                            AddLine('Page $' + IntToHex(Page, 2) + ': ' + BytesToHex(Response));

                          NfcTagSession.Invalidate;
                        end);
                    end;
                  ReadProc(4 * Page);
                end;

                NfcTagSession.Invalidate;
              end;
            end);
        end

        else
          AddLine('Unknown tag type');
      end;

    NfcTagSession.Invalidate;
  except
    on E: Exception do
      AddLine('Exception: ' + E.Message);
  end;
end;

procedure TFormMain.NfcError(Sender: TObject; const Error: TNfcError);
begin
  if Error.Code <> NFCReaderSessionInvalidationErrorUserCanceled then
  begin
    AddLine('Error event');
    ShowError(Error);
  end;
end;

procedure TFormMain.ButtonNdefClick(Sender: TObject);
begin
  Memo.Lines.Clear;
  try
    if not TNDefSession.ReadingAvailable then
      AddLine('NFC is not available on this device')
    else
    begin
      NDefSession.Free;
      NDefSession := TNDefSession.Create;
      NDefSession.AlertMessage := 'Hold your iPhone near the item.';
      NDefSession.OnActive := NfcActive;
      NDefSession.OnDetect := NDefDetect;
      NDefSession.OnError := NfcError;
      NDefSession.Scan;
    end;
  except
    on E: Exception do
      AddLine('Exception: ' + E.Message);
  end;
end;

procedure TFormMain.ButtonNfcTagClick(Sender: TObject);
begin
  Memo.Lines.Clear;
  try
    if not TNfcTagSession.ReadingAvailable then
      AddLine('NFC is not available on this device')
    else
    begin
      NfcTagSession.Free;
      NfcTagSession := TNfcTagSession.Create;
      NfcTagSession.AlertMessage := 'Hold your iPhone near the item.';
      NfcTagSession.OnActive := NfcActive;
      NfcTagSession.OnDetect := NfcDetect;
      NfcTagSession.OnError := NfcError;
      NfcTagSession.Scan;
    end;
  except
    on E: Exception do
      AddLine('Exception: ' + E.Message);
  end;
end;

procedure TFormMain.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  NDefSession.Free;
  NfcTagSession.Free;
end;

end.
